# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

from airflow.models import Connection
from airflow.operators.dummy_operator import DummyOperator

from marquez_airflow.extractors import StepMetadata
from marquez_airflow.version import VERSION as MARQUEZ_AIRFLOW_VERSION
from marquez_airflow.utils import (
    url_to_https,
    get_location,
    get_connection_uri,
    add_airflow_info_to
)

AIRFLOW_VERSION = '1.10.12'
AIRFLOW_CONN_ID = 'test_db'
AIRFLOW_CONN_URI = 'postgres://localhost:5432/testdb'


@mock.patch("marquez_airflow.utils._get_connection")
def test_get_connection_uri(mock_get_connection):
    mock_get_connection.return_value = Connection(
        conn_id=AIRFLOW_CONN_ID,
        uri=AIRFLOW_CONN_URI
    )
    assert get_connection_uri(AIRFLOW_CONN_ID) == \
        AIRFLOW_CONN_URI


def test_get_connection_uri_from_env():
    # Set the environment variable as AIRFLOW_CONN_<conn_id>
    os.environ[f"AIRFLOW_CONN_{AIRFLOW_CONN_ID.upper()}"] = AIRFLOW_CONN_URI
    assert get_connection_uri(AIRFLOW_CONN_ID) == AIRFLOW_CONN_URI


def test_add_airflow_info_to():
    task = DummyOperator(task_id='test.task')
    steps_metadata = [StepMetadata(name='test.task')]

    add_airflow_info_to(task, steps_metadata)
    for step_metadata in steps_metadata:
        assert step_metadata.context['airflow.operator'] == \
            'airflow.operators.dummy_operator.DummyOperator'
        assert step_metadata.context['airflow.version'] == AIRFLOW_VERSION
        assert step_metadata.context['airflow.task_info'] is not None
        assert step_metadata.context['marquez_airflow.version'] == \
            MARQUEZ_AIRFLOW_VERSION


def test_get_location_no_file_path():
    assert get_location(None) is None
    assert get_location("") is None


def test_url_to_https_no_url():
    assert url_to_https(None) is None
    assert url_to_https("") is None



import os
import logging
from marquez_client import Clients
from marquez_airflow.extractors import (Dataset, Source)
from marquez_client.models import JobType
from airflow.utils.state import State

_DAG_DEFAULT_OWNER = 'anonymous'
_DAG_DEFAULT_NAMESPACE = 'default'

log = logging.getLogger(__name__)


class Marquez:
    _marquez_client = None
    marquez_namespace = os.getenv('MARQUEZ_NAMESPACE',
                                  _DAG_DEFAULT_NAMESPACE)

    def get_or_create_marquez_client(self):
        if not self._marquez_client:
            self._marquez_client = Clients.new_write_only_client()
        return self._marquez_client

    def create_namespace(self):
        # TODO: Use 'anonymous' owner for now, but we may want to use
        # the 'owner' attribute defined via default_args for a DAG
        log.debug(
            f"Creating namespace '{self.marquez_namespace}' with "
            f"owner '{_DAG_DEFAULT_OWNER}'...")
        self.get_or_create_marquez_client() \
            .create_namespace(self.marquez_namespace, _DAG_DEFAULT_OWNER)

    def create_run(self, run_id, step, run_args, start_time, end_time) -> str:
        marquez_client = self.get_or_create_marquez_client()
        marquez_client.create_job_run(
            namespace_name=self.marquez_namespace,
            job_name=step.name,
            run_id=run_id,
            run_args=run_args,
            nominal_start_time=start_time,
            nominal_end_time=end_time)
        return run_id

    def create_datasets(self, datasets, marquez_job_run_id=None):
        client = self.get_or_create_marquez_client()
        for dataset in datasets:
            if isinstance(dataset, Dataset):
                self.create_source(dataset.source)
                # NOTE: The client expects a dict when capturing
                # fields for a dataset. Below we translate a field
                # object into a dict for compatibility. Work is currently
                # in progress to make this step unnecessary (see:
                # https://github.com/MarquezProject/marquez-python/pull/89)
                fields = []
                for field in dataset.fields:
                    fields.append({
                        'name': field.name,
                        'type': field.type,
                        'tags': field.tags,
                        'description': field.description
                    })
                client.create_dataset(
                    dataset_name=dataset.name,
                    dataset_type=dataset.type,
                    physical_name=dataset.name,
                    source_name=dataset.source.name,
                    namespace_name=self.marquez_namespace,
                    fields=fields,
                    run_id=marquez_job_run_id)

    def create_source(self, source):
        if isinstance(source, Source):
            client = self.get_or_create_marquez_client()
            client.create_source(
                source.name,
                source.type,
                source.connection_url)

    def create_job(self, step, location, description, state=None, run_id=None):
        # Create a job to ensure it exists
        self.get_or_create_marquez_client().create_job(
            namespace_name=self.marquez_namespace,
            job_name=step.name,
            job_type=JobType.BATCH,
            location=step.location or location,
            input_dataset=self._register_dataset(step.inputs),
            output_dataset=self._register_dataset(step.outputs, run_id)
            if state in {State.SUCCESS, State.SKIPPED}
            else self._register_dataset(step.outputs),
            context=step.context or {},
            description=description,
            run_id=run_id)

    def _register_dataset(self, step_dataset, run_id=None):
        inputs = None
        if step_dataset:
            self.create_datasets(step_dataset, run_id)
            inputs = self._to_dataset_ids(step_dataset)
        return inputs

    def complete_run(self, run_id):
        self.get_or_create_marquez_client(). \
            mark_job_run_as_completed(run_id=run_id)

    def fail_run(self, run_id):
        self.get_or_create_marquez_client().mark_job_run_as_failed(
            run_id=run_id)

    def start_run(self, run_id, start):
        self.get_or_create_marquez_client() \
            .mark_job_run_as_started(run_id, start)

    def _to_dataset_ids(self, datasets):
        if not datasets:
            return None

        return list(map(lambda ds: {
            'namespace': self.marquez_namespace,
            'name': ds.name
        }, datasets))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import closing
from typing import Optional

from airflow.hooks.postgres_hook import PostgresHook
from airflow.operators.postgres_operator import PostgresOperator

from marquez_airflow.models import DbTableSchema, DbColumn
from marquez_airflow.utils import get_connection_uri
from marquez_airflow.extractors.sql.experimental import SqlMeta
from marquez_airflow.extractors.sql.experimental.parser import SqlParser
from marquez_airflow.extractors import (
    BaseExtractor,
    StepMetadata,
    Source,
    Dataset
)

_TABLE_SCHEMA = 0
_TABLE_NAME = 1
_COLUMN_NAME = 2
_ORDINAL_POSITION = 3
# Use 'udt_name' which is the underlying type of column
# (ex: int4, timestamp, varchar, etc)
_UDT_NAME = 4


class PostgresExtractor(BaseExtractor):
    operator_class = PostgresOperator

    def __init__(self, operator):
        super().__init__(operator)

    def extract(self) -> [StepMetadata]:
        # (1) Parse sql statement to obtain input / output tables.
        sql_meta: SqlMeta = SqlParser.parse(self.operator.sql)

        # (2) Default all inputs / outputs to current connection.
        # NOTE: We'll want to look into adding support for the `database`
        # property that is used to override the one defined in the connection.
        conn_id = self.operator.postgres_conn_id
        source = Source(
            type='POSTGRESQL',
            name=conn_id,
            connection_url=get_connection_uri(conn_id))

        # (3) Map input / output tables to dataset objects with source set
        # as the current connection. We need to also fetch the schema for the
        # input tables to format the dataset name as:
        # {schema_name}.{table_name}
        inputs = [
            Dataset.from_table(
                source=source,
                table_name=in_table_schema.table_name,
                schema_name=in_table_schema.schema_name
            ) for in_table_schema in self._get_table_schemas(
                sql_meta.in_tables
            )
        ]
        outputs = [
            Dataset.from_table_schema(
                source=source,
                table_schema=out_table_schema
            ) for out_table_schema in self._get_table_schemas(
                sql_meta.out_tables
            )
        ]

        return [StepMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=inputs,
            outputs=outputs,
            context={
                'sql': self.operator.sql
            }
        )]

    def _get_table_schemas(self, tables: [str]) -> [DbTableSchema]:
        # Avoid querying postgres by returning an empty array
        # if no tables have been provided.
        if not tables:
            return []

        # Keeps tack of the schema by table.
        schemas_by_table = {}

        hook = PostgresHook(
            postgres_conn_id=self.operator.postgres_conn_id,
            schema=self.operator.database
        )
        with closing(hook.get_conn()) as conn:
            with closing(conn.cursor()) as cursor:
                table_names = ",".join(map(lambda table: f"'{table}'", tables))
                cursor.execute(
                    f"""
                    SELECT table_schema,
                           table_name,
                           column_name,
                           ordinal_position,
                           udt_name
                      FROM information_schema.columns
                     WHERE table_name IN ({table_names});
                    """
                )
                for row in cursor.fetchall():
                    table_schema_name: str = row[_TABLE_SCHEMA]
                    table_name: str = row[_TABLE_NAME]
                    table_column: DbColumn = DbColumn(
                        name=row[_COLUMN_NAME],
                        type=row[_UDT_NAME],
                        ordinal_position=row[_ORDINAL_POSITION]
                    )

                    # Attempt to get table schema
                    table_key: str = f"{table_schema_name}.{table_name}"
                    table_schema: Optional[DbTableSchema] = \
                        schemas_by_table.get(table_key)

                    if table_schema:
                        # Add column to existing table schema.
                        schemas_by_table[table_key]. \
                            columns.append(table_column)
                    else:
                        # Create new table schema with column.
                        schemas_by_table[table_key] = DbTableSchema(
                            schema_name=table_schema_name,
                            table_name=table_name,
                            columns=[table_column]
                        )

        return list(schemas_by_table.values())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from uuid import uuid4

import airflow.models
import time
from airflow import LoggingMixin
from airflow.contrib.operators.bigquery_operator import BigQueryOperator
from airflow.operators.postgres_operator import PostgresOperator
from airflow.utils.state import State

from marquez_airflow.extractors import StepMetadata
from marquez_airflow.extractors.bigquery_extractor import BigQueryExtractor
from marquez_airflow.extractors.postgres_extractor import PostgresExtractor
from marquez_airflow.utils import (
    JobIdMapping,
    get_location,
    add_airflow_info_to,
    DagUtils
)

from marquez_airflow.marquez import Marquez

log = logging.getLogger(__name__)


class DAG(airflow.models.DAG, LoggingMixin):
    _job_id_mapping = None
    _marquez = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._job_id_mapping = JobIdMapping()
        self._marquez = Marquez()
        # TODO: Manually define operator->extractor mappings for now,
        # but we'll want to encapsulate this logic in an 'Extractors' class
        # with more convenient methods (ex: 'Extractors.extractor_for_task()')
        self._extractors = {
            PostgresOperator: PostgresExtractor,
            BigQueryOperator: BigQueryExtractor
            # Append new extractors here
        }
        self.log.debug(
            f"DAG successfully created with extractors: {self._extractors}"
        )

    def create_dagrun(self, *args, **kwargs):
        # run Airflow's create_dagrun() first
        dagrun = super(DAG, self).create_dagrun(*args, **kwargs)

        create_dag_start_ms = self._now_ms()
        try:
            self._marquez.create_namespace()
            self._register_dagrun(
                dagrun,
                DagUtils.get_execution_date(**kwargs),
                DagUtils.get_run_args(**kwargs)
            )
        except Exception as e:
            self.log.error(
                f'Failed to record metadata: {e} '
                f'{self._timed_log_message(create_dag_start_ms)}',
                exc_info=True)

        return dagrun

    def _register_dagrun(self, dagrun, execution_date, run_args):
        self.log.debug(f"self.task_dict: {self.task_dict}")
        # Register each task in the DAG
        for task_id, task in self.task_dict.items():
            t = self._now_ms()
            try:
                steps = self._extract_metadata(dagrun, task)
                [self._marquez.create_job(
                    step, self._get_location(task), self.description)
                    for step in steps]
                marquez_jobrun_ids = [self._marquez.create_run(
                    self.new_run_id(),
                    step,
                    run_args,
                    DagUtils.get_start_time(execution_date),
                    DagUtils.get_end_time(
                        execution_date,
                        self.following_schedule(execution_date))
                ) for step in steps]
                self._job_id_mapping.set(
                    self._marquez_job_name(self.dag_id, task.task_id),
                    dagrun.run_id,
                    marquez_jobrun_ids)
            except Exception as e:
                self.log.error(
                    f'Failed to record task {task_id}: {e} '
                    f'{self._timed_log_message(t)}',
                    exc_info=True)

    def handle_callback(self, *args, **kwargs):
        self.log.debug(f"handle_callback({args}, {kwargs})")
        try:
            dagrun = args[0]
            self.log.debug(f"handle_callback() dagrun : {dagrun}")
            self._marquez.create_namespace()
            self._report_task_instances(
                dagrun,
                DagUtils.get_run_args(**kwargs),
                kwargs.get('session')
            )
        except Exception as e:
            self.log.error(
                f'Failed to record dagrun callback: {e} '
                f'dag_id={self.dag_id}',
                exc_info=True)

        return super().handle_callback(*args)

    def _report_task_instances(self, dagrun, run_args, session):
        task_instances = dagrun.get_task_instances()
        for ti in task_instances:
            try:
                self._report_task_instance(ti, dagrun, run_args, session)
            except Exception as e:
                self.log.error(
                    f'Failed to record task instance: {e} '
                    f'dag_id={self.dag_id}',
                    exc_info=True)

    def _report_task_instance(self, ti, dagrun, run_args, session):
        task = self.get_task(ti.task_id)
        run_ids = self._job_id_mapping.pop(
            self._marquez_job_name_from_ti(ti), dagrun.run_id, session)
        steps = self._extract_metadata(dagrun, task, ti)

        # Note: run_ids could be missing if it was removed from airflow
        # or the job could not be registered.
        if not run_ids:
            [self._marquez.create_job(
                step, self._get_location(task), self.description)
             for step in steps]
            run_ids = [self._marquez.create_run(
                self.new_run_id(),
                step,
                run_args,
                DagUtils.to_iso_8601(ti.start_date),
                DagUtils.to_iso_8601(ti.end_date)
            ) for step in steps]
            if not run_ids:
                self.log.warn('Could not emit lineage')

        for step in steps:
            for run_id in run_ids:
                self._marquez.create_job(
                    step, self._get_location(task), self.description,
                    ti.state, run_id)
                self._marquez.start_run(
                    run_id,
                    DagUtils.to_iso_8601(ti.start_date))

                self.log.debug(f'Setting task state: {ti.state}'
                               f' for {ti.task_id}')
                if ti.state in {State.SUCCESS, State.SKIPPED}:
                    self._marquez.complete_run(run_id)
                else:
                    self._marquez.fail_run(run_id)

    def _extract_metadata(self, dagrun, task, ti=None):
        extractor = self._get_extractor(task)
        task_info = f'task_type={task.__class__.__name__} ' \
            f'airflow_dag_id={self.dag_id} ' \
            f'task_id={task.task_id} ' \
            f'airflow_run_id={dagrun.run_id} '
        if extractor:
            try:
                self.log.debug(
                    f'Using extractor {extractor.__name__} {task_info}')
                steps = self._extract(extractor, task, ti)

                return add_airflow_info_to(
                    task,
                    steps
                )
            except Exception as e:
                self.log.error(
                    f'Failed to extract metadata {e} {task_info}',
                    exc_info=True)
        else:
            self.log.warning(
                f'Unable to find an extractor. {task_info}')

        return add_airflow_info_to(
            task,
            [StepMetadata(name=self._marquez_job_name(
                self.dag_id, task.task_id))]
        )

    def _extract(self, extractor, task, ti):
        if ti:
            steps = extractor(task).extract_on_complete(ti)
            if steps:
                return steps

        return extractor(task).extract()

    def _get_extractor(self, task):
        extractor = self._extractors.get(task.__class__)
        log.debug(f'extractor for {task.__class__} is {extractor}')
        return extractor

    def _timed_log_message(self, start_time):
        return f'airflow_dag_id={self.dag_id} ' \
            f'duration_ms={(self._now_ms() - start_time)}'

    def new_run_id(self) -> str:
        return str(uuid4())

    @staticmethod
    def _now_ms():
        return int(round(time.time() * 1000))

    @staticmethod
    def _get_location(task):
        try:
            if hasattr(task, 'file_path') and task.file_path:
                return get_location(task.file_path)
            else:
                return get_location(task.dag.fileloc)
        except Exception:
            log.warning(f"Failed to get location for task '{task.task_id}'.",
                        exc_info=True)
            return None

    @staticmethod
    def _marquez_job_name_from_ti(ti):
        return DAG._marquez_job_name(ti.dag_id, ti.task_id)

    @staticmethod
    def _marquez_job_name(dag_id, task_id):
        return f'{dag_id}.{task_id}'



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from abc import ABC, abstractmethod

from airflow import LoggingMixin
from airflow.models import BaseOperator

from marquez_airflow.models import DbTableSchema, DbColumn

from marquez_client.models import DatasetType


class Source:
    name = None
    connection_url = None
    type = None

    def __init__(self, name, type, connection_url):
        self.name = name
        self.type = type
        self.connection_url = connection_url

    def __eq__(self, other):
        return self.name == other.name and \
               self.type == other.type and \
               self.connection_url == other.connection_url

    def __repr__(self):
        return f"Source({self.name!r},{self.type!r},{self.connection_url!r})"


class Field:
    def __init__(self, name: str, type: str,
                 tags: List[str] = [], description: str = None):
        self.name = name
        self.type = type
        self.tags = tags
        self.description = description

    @staticmethod
    def from_column(column: DbColumn):
        return Field(
            name=column.name,
            type=column.type,
            description=column.description
        )

    def __eq__(self, other):
        return self.name == other.name and \
               self.type == other.type and \
               self.tags == other.tags and \
               self.description == other.description

    def __repr__(self):
        return f"Field({self.name!r},{self.type!r}, \
                       {self.tags!r},{self.description!r})"


class Dataset:
    def __init__(self, source: Source, name: str, type: DatasetType,
                 fields: List[Field] = [], description=None):
        self.source = source
        self.name = name
        self.type = type
        self.fields = fields
        self.description = description

    @staticmethod
    def from_table(source: Source, table_name: str,
                   schema_name: str = None):
        return Dataset(
            type=DatasetType.DB_TABLE,
            name=Dataset._to_name(
                schema_name=schema_name,
                table_name=table_name
            ),
            source=source
        )

    @staticmethod
    def from_table_schema(source: Source, table_schema: DbTableSchema):
        return Dataset(
            type=DatasetType.DB_TABLE,
            name=Dataset._to_name(
                schema_name=table_schema.schema_name,
                table_name=table_schema.table_name
            ),
            source=source,
            fields=[
                # We want to maintain column order using ordinal position.
                Field.from_column(column) for column in sorted(
                    table_schema.columns, key=lambda x: x.ordinal_position
                )
            ]
        )

    @staticmethod
    def _to_name(table_name: str, schema_name: str = None):
        # Prefix the table name with the schema name using
        # the format: {table_schema}.{table_name}.
        return f"{schema_name}.{table_name}" if schema_name else table_name

    def __eq__(self, other):
        return self.source == other.source and \
               self.name == other.name and \
               self.type == other.type and \
               self.fields == other.fields and \
               self.description == other.description

    def __repr__(self):
        return f"Dataset({self.source!r},{self.name!r}, \
                         {self.type!r},{self.fields!r},{self.description!r})"


class StepMetadata:
    # TODO: Define a common way across extractors to build the
    # job name for an operator
    name = None
    location = None
    inputs = []
    outputs = []
    context = {}

    def __init__(self, name, location=None, inputs=None, outputs=None,
                 context=None):
        self.name = name
        self.location = location
        if inputs:
            self.inputs = inputs
        if outputs:
            self.outputs = outputs
        if context:
            self.context = context

    def __repr__(self):
        return "name: {}\t inputs: {} \t outputs: {}".format(
            self.name,
            ','.join([str(i) for i in self.inputs]),
            ','.join([str(o) for o in self.outputs]))


class BaseExtractor(ABC, LoggingMixin):
    operator: BaseOperator = None
    operator_class = None

    def __init__(self, operator):
        self.operator = operator

    @classmethod
    def get_operator_class(cls):
        return cls.operator_class

    def validate(self):
        # TODO: maybe we should also enforce the module
        assert (self.operator_class is not None and
                self.operator.__class__ == self.operator_class)

    # TODO: Only return a single StepMetadata object.
    @abstractmethod
    def extract(self) -> [StepMetadata]:
        pass

    def extract_on_complete(self, task_instance) -> [StepMetadata]:
        # TODO: This method allows for the partial updating of task
        # metadata on completion. Marquez currently doesn't support
        # partial updates within the context of a DAG run, but this feature
        # will soon be supported:
        # https://github.com/MarquezProject/marquez/issues/816
        #
        # Also, we'll want to revisit the metadata extraction flow,
        # but for now, return an empty set as the default behavior
        # as not all extractors need to handle partial metadata updates.
        return self.extract()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import mock
import logging

from airflow.models import (TaskInstance, DagRun)
from airflow.operators.dummy_operator import DummyOperator
from airflow.utils.decorators import apply_defaults
from airflow.utils.db import provide_session
from airflow.utils.dates import days_ago
from airflow.utils import timezone
from airflow.utils.state import State

from marquez_client.models import JobType, DatasetType

from marquez_airflow import DAG
from marquez_airflow.extractors import (
    BaseExtractor, StepMetadata, Source, Dataset
)
from marquez_airflow.models import DbTableSchema, DbColumn
from marquez_airflow.utils import get_location, get_job_name

from uuid import UUID

log = logging.getLogger(__name__)

NO_INPUTS = []
NO_OUTPUTS = []

DEFAULT_DATE = timezone.datetime(2016, 1, 1)

DAG_ID = 'test_dag'
DAG_RUN_ID = 'test_run_id_for_task_completed_and_failed'
DAG_RUN_ARGS = {'external_trigger': False}
# TODO: check with a different namespace and owner
DAG_NAMESPACE = 'default'
DAG_OWNER = 'anonymous'
DAG_DESCRIPTION = \
    'A simple DAG to test the marquez.DAG metadata extraction flow.'

DAG_DEFAULT_ARGS = {
    'owner': DAG_OWNER,
    'depends_on_past': False,
    'start_date': days_ago(1),
    'email_on_failure': False,
    'email_on_retry': False,
    'email': ['owner@test.com']
}

TASK_ID_COMPLETED = 'test_task_completed'
TASK_ID_FAILED = 'test_task_failed'


@pytest.fixture
@provide_session
def clear_db_airflow_dags(session=None):
    session.query(DagRun).delete()
    session.query(TaskInstance).delete()


@provide_session
def test_new_run_id(clear_db_airflow_dags, session=None):
    dag = DAG(
        DAG_ID,
        schedule_interval='@daily',
        default_args=DAG_DEFAULT_ARGS,
        description=DAG_DESCRIPTION
    )
    run_id = dag.new_run_id()
    assert UUID(run_id).version == 4


# tests a simple workflow with default extraction mechanism
@mock.patch('marquez_airflow.DAG.new_run_id')
@mock.patch('marquez_airflow.marquez.Marquez.get_or_create_marquez_client')
@provide_session
def test_marquez_dag(mock_get_or_create_marquez_client, mock_uuid,
                     clear_db_airflow_dags, session=None):

    dag = DAG(
        DAG_ID,
        schedule_interval='@daily',
        default_args=DAG_DEFAULT_ARGS,
        description=DAG_DESCRIPTION
    )
    # (1) Mock the marquez client method calls
    mock_marquez_client = mock.Mock()
    mock_get_or_create_marquez_client.return_value = mock_marquez_client
    run_id_completed = "my-test_marquez_dag-uuid-completed"
    run_id_failed = "my-test_marquez_dag-uuid-failed"
    mock_uuid.side_effect = [run_id_completed, run_id_failed]

    # (2) Add task that will be marked as completed
    task_will_complete = DummyOperator(
        task_id=TASK_ID_COMPLETED,
        dag=dag
    )
    completed_task_location = get_location(task_will_complete.dag.fileloc)

    # (3) Add task that will be marked as failed
    task_will_fail = DummyOperator(
        task_id=TASK_ID_FAILED,
        dag=dag
    )
    failed_task_location = get_location(task_will_complete.dag.fileloc)

    # (4) Create DAG run and mark as running
    dagrun = dag.create_dagrun(
        run_id=DAG_RUN_ID,
        execution_date=DEFAULT_DATE,
        state=State.RUNNING)

    # Assert namespace meta call
    mock_marquez_client.create_namespace.assert_called_once_with(DAG_NAMESPACE,
                                                                 DAG_OWNER)

    # Assert source and dataset meta calls
    mock_marquez_client.create_source.assert_not_called()
    mock_marquez_client.create_dataset.assert_not_called()

    # Assert job meta calls
    create_job_calls = [
        mock.call(
            job_name=f"{DAG_ID}.{TASK_ID_COMPLETED}",
            job_type=JobType.BATCH,
            location=completed_task_location,
            input_dataset=None,
            output_dataset=None,
            context=mock.ANY,
            description=DAG_DESCRIPTION,
            namespace_name=DAG_NAMESPACE,
            run_id=None
        ),
        mock.call(
            job_name=f"{DAG_ID}.{TASK_ID_FAILED}",
            job_type=JobType.BATCH,
            location=failed_task_location,
            input_dataset=None,
            output_dataset=None,
            context=mock.ANY,
            description=DAG_DESCRIPTION,
            namespace_name=DAG_NAMESPACE,
            run_id=None
        )
    ]
    log.info(
        f"{ [name for name, args, kwargs in mock_marquez_client.mock_calls]}")
    mock_marquez_client.create_job.assert_has_calls(create_job_calls)

    # Assert job run meta calls
    create_job_run_calls = [
        mock.call(
            job_name=f"{DAG_ID}.{TASK_ID_COMPLETED}",
            run_id=mock.ANY,
            run_args=DAG_RUN_ARGS,
            nominal_start_time=mock.ANY,
            nominal_end_time=mock.ANY,
            namespace_name=DAG_NAMESPACE
        ),
        mock.call(
            job_name=f"{DAG_ID}.{TASK_ID_FAILED}",
            run_id=mock.ANY,
            run_args=DAG_RUN_ARGS,
            nominal_start_time=mock.ANY,
            nominal_end_time=mock.ANY,
            namespace_name=DAG_NAMESPACE
        )
    ]
    mock_marquez_client.create_job_run.assert_has_calls(create_job_run_calls)

    # (5) Start task that will be marked as completed
    task_will_complete.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

    # (6) Start task that will be marked as failed
    ti1 = TaskInstance(task=task_will_fail, execution_date=DEFAULT_DATE)
    ti1.state = State.FAILED
    session.add(ti1)
    session.commit()

    dag.handle_callback(dagrun, success=True, session=session)

    # Assert start run meta calls
    start_job_run_calls = [
        mock.call(run_id_completed, mock.ANY),
        mock.call(run_id_failed, mock.ANY)
    ]
    mock_marquez_client.mark_job_run_as_started.assert_has_calls(
        start_job_run_calls
    )

    mock_marquez_client.mark_job_run_as_completed.assert_called_once_with(
        run_id=run_id_completed
    )

    # When a task run completes, the task outputs are also updated in order
    # to link a job version (=task version) to a dataset version.
    # Using a DummyOperator, no outputs exists, so assert that the create
    # dataset call is not invoked.
    mock_marquez_client.create_dataset.assert_not_called()

    dag.handle_callback(dagrun, success=False, session=session)
    mock_marquez_client.mark_job_run_as_failed.assert_called_once_with(
        run_id=run_id_failed
    )

    # Assert an attempt to version the outputs of a task is not made when
    # a task fails
    mock_marquez_client.create_dataset.assert_not_called()


class TestFixtureDummyOperator(DummyOperator):

    @apply_defaults
    def __init__(self, *args, **kwargs):
        super(TestFixtureDummyOperator, self).__init__(*args, **kwargs)


class TestFixtureDummyExtractor(BaseExtractor):
    operator_class = TestFixtureDummyOperator
    source = Source(
        type="DummySource",
        name="dummy_source_name",
        connection_url="http://dummy/source/url")

    def __init__(self, operator):
        super().__init__(operator)

    def extract(self) -> [StepMetadata]:
        inputs = [
            Dataset.from_table(self.source, "extract_input1")
        ]
        outputs = [
            Dataset.from_table(self.source, "extract_output1")
        ]
        return [StepMetadata(
            name=get_job_name(task=self.operator),
            inputs=inputs,
            outputs=outputs,
            context={
                "extract": "extract"
            }
        )]

    def extract_on_complete(self, task_instance) -> [StepMetadata]:
        return []


class TestFixtureDummyExtractorOnComplete(BaseExtractor):
    operator_class = TestFixtureDummyOperator
    source = Source(
        type="DummySource",
        name="dummy_source_name",
        connection_url="http://dummy/source/url")

    def __init__(self, operator):
        super().__init__(operator)

    def extract(self) -> [StepMetadata]:
        return []

    def extract_on_complete(self, task_instance) -> [StepMetadata]:
        inputs = [
            Dataset.from_table_schema(self.source, DbTableSchema(
                schema_name='schema',
                table_name='extract_on_complete_input1',
                columns=[DbColumn(
                    name='field1',
                    type='text',
                    description='',
                    ordinal_position=1
                ),
                    DbColumn(
                    name='field2',
                    type='text',
                    description='',
                    ordinal_position=2
                )]
            ))
        ]
        outputs = [
            Dataset.from_table(self.source, "extract_on_complete_output1")
        ]
        return [StepMetadata(
            name=get_job_name(task=self.operator),
            inputs=inputs,
            outputs=outputs,
            context={
                "extract_on_complete": "extract_on_complete"
            }
        )]


# test the lifecycle including with extractors
@mock.patch('marquez_airflow.DAG.new_run_id')
@mock.patch('marquez_airflow.marquez.Marquez.get_or_create_marquez_client')
@provide_session
def test_marquez_dag_with_extractor(mock_get_or_create_marquez_client,
                                    mock_uuid,
                                    clear_db_airflow_dags,
                                    session=None):

    # --- test setup
    dag_id = 'test_marquez_dag_with_extractor'
    dag = DAG(
        dag_id,
        schedule_interval='@daily',
        default_args=DAG_DEFAULT_ARGS,
        description=DAG_DESCRIPTION
    )

    run_id = "my-test-uuid"
    mock_uuid.side_effect = [run_id]
    # Mock the marquez client method calls
    mock_marquez_client = mock.Mock()
    mock_get_or_create_marquez_client.return_value = mock_marquez_client

    # Add task that will be marked as completed
    task_will_complete = TestFixtureDummyOperator(
        task_id=TASK_ID_COMPLETED,
        dag=dag
    )
    completed_task_location = get_location(task_will_complete.dag.fileloc)

    # Add the dummy extractor to the list for the task above
    dag._extractors[task_will_complete.__class__] = TestFixtureDummyExtractor

    # --- pretend run the DAG

    # Create DAG run and mark as running
    dagrun = dag.create_dagrun(
        run_id='test_marquez_dag_with_extractor_run_id',
        execution_date=DEFAULT_DATE,
        state=State.RUNNING)

    # --- Asserts that the job starting triggers metadata updates

    # Namespace created
    mock_marquez_client.create_namespace.assert_called_once_with(DAG_NAMESPACE,
                                                                 DAG_OWNER)

    # Datasets are updated
    mock_marquez_client.create_source.assert_called_with(
        'dummy_source_name',
        'DummySource',
        'http://dummy/source/url'
    )
    mock_marquez_client.create_dataset.assert_has_calls([
        mock.call(
            dataset_name='extract_input1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='extract_input1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=[],
            run_id=None
        ),
        mock.call(
            dataset_name='extract_output1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='extract_output1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=[],
            run_id=None
        )
    ])

    # job is updated
    mock_marquez_client.create_job.assert_called_once_with(
        job_name=f"{dag_id}.{TASK_ID_COMPLETED}",
        job_type=JobType.BATCH,
        location=completed_task_location,
        input_dataset=[{'namespace': 'default', 'name': 'extract_input1'}],
        output_dataset=[{'namespace': 'default', 'name': 'extract_output1'}],
        context=mock.ANY,
        description=DAG_DESCRIPTION,
        namespace_name=DAG_NAMESPACE,
        run_id=None
    )
    assert mock_marquez_client.create_job.mock_calls[0].\
        kwargs['context'].get('extract') == 'extract'

    # run is created
    mock_marquez_client.create_job_run.assert_called_once_with(
        job_name=f"{dag_id}.{TASK_ID_COMPLETED}",
        run_id=run_id,
        run_args=DAG_RUN_ARGS,
        nominal_start_time=mock.ANY,
        nominal_end_time=mock.ANY,
        namespace_name=DAG_NAMESPACE
    )

    log.info("Marquez client calls when starting:")
    for call in mock_marquez_client.mock_calls:
        log.info(call)

    assert [name for name, args, kwargs in mock_marquez_client.mock_calls] == [
        'create_namespace',
        'create_source',
        'create_dataset',
        'create_source',
        'create_dataset',
        'create_job',
        'create_job_run'
    ]
    mock_marquez_client.reset_mock()

    # --- Pretend complete the task
    task_will_complete.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

    dag.handle_callback(dagrun, success=True, session=session)

    # run is started
    mock_marquez_client.mark_job_run_as_started.assert_called_once_with(
        run_id, mock.ANY
    )

    # --- Assert that the right marquez calls are done

    # job is updated before completion
    mock_marquez_client.create_job.assert_has_calls([
        mock.call(
            namespace_name=DAG_NAMESPACE,
            job_name=f"{dag_id}.{TASK_ID_COMPLETED}",
            job_type=JobType.BATCH,
            location=completed_task_location,
            input_dataset=[
                {'namespace': 'default', 'name': 'extract_input1'}
            ],
            output_dataset=[
                {'namespace': 'default', 'name': 'extract_output1'}
            ],
            context=mock.ANY,
            description=DAG_DESCRIPTION,
            run_id=run_id
        )
    ])

    assert mock_marquez_client.create_job.mock_calls[0].\
        kwargs['context'].get('extract') == 'extract'

    mock_marquez_client.mark_job_run_as_completed.assert_called_once_with(
        run_id=run_id
    )

    # When a task run completes, the task outputs are also updated in order
    # to link a job version (=task version) to a dataset version.
    mock_marquez_client.create_dataset.assert_has_calls([
        mock.call(
            dataset_name='extract_input1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='extract_input1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=[],
            run_id=None
        ),
        mock.call(
            dataset_name='extract_output1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='extract_output1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=[],
            run_id=run_id
        )
    ])

    log.info("Marquez client calls when completing:")
    for call in mock_marquez_client.mock_calls:
        log.info(call)
    assert [name for name, args, kwargs in mock_marquez_client.mock_calls] == [
        'create_namespace',
        'create_source',
        'create_dataset',
        'create_source',
        'create_dataset',
        'create_job',
        'mark_job_run_as_started',
        'mark_job_run_as_completed'
    ]


@mock.patch('marquez_airflow.DAG.new_run_id')
@mock.patch('marquez_airflow.marquez.Marquez.get_or_create_marquez_client')
@provide_session
def test_marquez_dag_with_extract_on_complete(
        mock_get_or_create_marquez_client,
        mock_uuid,
        clear_db_airflow_dags,
        session=None):

    # --- test setup
    dag_id = 'test_marquez_dag_with_extractor'
    dag = DAG(
        dag_id,
        schedule_interval='@daily',
        default_args=DAG_DEFAULT_ARGS,
        description=DAG_DESCRIPTION
    )

    run_id = "my-test-uuid"
    mock_uuid.side_effect = [run_id]
    # Mock the marquez client method calls
    mock_marquez_client = mock.Mock()
    mock_get_or_create_marquez_client.return_value = mock_marquez_client

    # Add task that will be marked as completed
    task_will_complete = TestFixtureDummyOperator(
        task_id=TASK_ID_COMPLETED,
        dag=dag
    )
    completed_task_location = get_location(task_will_complete.dag.fileloc)

    # Add the dummy extractor to the list for the task above
    dag._extractors[task_will_complete.__class__] = \
        TestFixtureDummyExtractorOnComplete

    # Create DAG run and mark as running
    dagrun = dag.create_dagrun(
        run_id='test_marquez_dag_with_extractor_run_id',
        execution_date=DEFAULT_DATE,
        state=State.RUNNING)

    # Namespace created
    mock_marquez_client.create_namespace.assert_called_once_with(DAG_NAMESPACE,
                                                                 DAG_OWNER)

    log.info("Marquez client calls when starting:")
    for call in mock_marquez_client.mock_calls:
        log.info(call)

    assert [name for name, args, kwargs in mock_marquez_client.mock_calls] == [
        'create_namespace'
    ]
    mock_marquez_client.reset_mock()

    # --- Pretend complete the task
    task_will_complete.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

    dag.handle_callback(dagrun, success=True, session=session)

    # Datasets are updated
    mock_marquez_client.create_source.assert_called_with(
        'dummy_source_name',
        'DummySource',
        'http://dummy/source/url'
    )
    # Datasets get called twice, once to reenact the _begin_run_flow
    #  and then again at _end_run_flow w/ the run id appended for
    #  the output dataset
    mock_marquez_client.create_dataset.assert_has_calls([
        mock.call(
            dataset_name='schema.extract_on_complete_input1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='schema.extract_on_complete_input1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=mock.ANY,
            run_id=None
        ),
        mock.call(
            dataset_name='extract_on_complete_output1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='extract_on_complete_output1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=[],
            run_id=None
        ),
        mock.call(
            dataset_name='schema.extract_on_complete_input1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='schema.extract_on_complete_input1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=mock.ANY,
            run_id=None
        ),
        mock.call(
            dataset_name='extract_on_complete_output1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='extract_on_complete_output1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=[],
            run_id='my-test-uuid'
        )
    ])

    # job is updated
    mock_marquez_client.create_job.assert_has_calls([
        mock.call(
            job_name=f"{dag_id}.{TASK_ID_COMPLETED}",
            job_type=JobType.BATCH,
            location=completed_task_location,
            input_dataset=[{'namespace': 'default',
                            'name': 'schema.extract_on_complete_input1'}],
            output_dataset=[{'namespace': 'default',
                             'name': 'extract_on_complete_output1'}],
            context=mock.ANY,
            description=DAG_DESCRIPTION,
            namespace_name=DAG_NAMESPACE,
            run_id=None
        ),
        mock.call(
            job_name=f"{dag_id}.{TASK_ID_COMPLETED}",
            job_type=JobType.BATCH,
            location=completed_task_location,
            input_dataset=[{'namespace': 'default',
                            'name': 'schema.extract_on_complete_input1'}],
            output_dataset=[{'namespace': 'default',
                             'name': 'extract_on_complete_output1'}],
            context=mock.ANY,
            description=DAG_DESCRIPTION,
            namespace_name=DAG_NAMESPACE,
            run_id='my-test-uuid'
        )
    ])
    assert mock_marquez_client.create_job.mock_calls[0].\
        kwargs['context'].get('extract_on_complete') == 'extract_on_complete'

    # run is created
    mock_marquez_client.create_job_run.assert_called_once_with(
        job_name=f"{dag_id}.{TASK_ID_COMPLETED}",
        run_id=run_id,
        run_args=DAG_RUN_ARGS,
        nominal_start_time=mock.ANY,
        nominal_end_time=mock.ANY,
        namespace_name=DAG_NAMESPACE
    )

    # run is started
    mock_marquez_client.mark_job_run_as_started.assert_called_once_with(
        run_id, mock.ANY
    )

    # --- Assert that the right marquez calls are done

    # job is updated before completion
    mock_marquez_client.create_job.assert_has_calls([
        mock.call(
            namespace_name=DAG_NAMESPACE,
            job_name=f"{dag_id}.{TASK_ID_COMPLETED}",
            job_type=JobType.BATCH,
            location=completed_task_location,
            input_dataset=[
                {'namespace': 'default',
                    'name': 'schema.extract_on_complete_input1'}
            ],
            output_dataset=[
                {'namespace': 'default', 'name': 'extract_on_complete_output1'}
            ],
            context=mock.ANY,
            description=DAG_DESCRIPTION,
            run_id=run_id
        )
    ])

    assert mock_marquez_client.create_job.mock_calls[0].\
        kwargs['context'].get('extract_on_complete') == 'extract_on_complete'

    mock_marquez_client.mark_job_run_as_completed.assert_called_once_with(
        run_id=run_id
    )

    # When a task run completes, the task outputs are also updated in order
    # to link a job version (=task version) to a dataset version.
    mock_marquez_client.create_dataset.assert_has_calls([
        mock.call(
            dataset_name='schema.extract_on_complete_input1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='schema.extract_on_complete_input1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=mock.ANY,
            run_id=None
        ),
        mock.call(
            dataset_name='extract_on_complete_output1',
            dataset_type=DatasetType.DB_TABLE,
            physical_name='extract_on_complete_output1',
            source_name='dummy_source_name',
            namespace_name=DAG_NAMESPACE,
            fields=[],
            run_id=run_id
        )
    ])

    log.info("Marquez client calls when completing:")
    for call in mock_marquez_client.mock_calls:
        log.info(call)
    assert [name for name, args, kwargs in mock_marquez_client.mock_calls] == [
        'create_namespace',
        'create_source',
        'create_dataset',
        'create_source',
        'create_dataset',
        'create_job',
        'create_job_run',
        'create_source',
        'create_dataset',
        'create_source',
        'create_dataset',
        'create_job',
        'mark_job_run_as_started',
        'mark_job_run_as_completed'
    ]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
from unittest.mock import create_autospec

from marquez_client import MarquezClient

log = logging.getLogger(__name__)


def get_mock_marquez_client():
    uuid_generator = generate_uuid()

    def create_job_run(*args, **kwargs):
        assert len(args) == 1
        log.debug("create_job_run()")
        return {"runId": str(next(uuid_generator))}

    def create_source(*args, **kwargs):
        assert len(args) >= 2
        name = args[0]
        # type = args[1] # commenting out as not used for now
        # connection_url = args[2]
        # source_type = connection_url.split(':')[1]
        return {'name': name}

    def create_dataset(*args, **kwargs):
        assert len(args) >= 2
        dataset_name = args[0]
        return {'name': dataset_name}

    client = create_autospec(MarquezClient)

    client.create_job_run.side_effect = create_job_run
    client.create_source.side_effect = create_source
    client.create_dataset.side_effect = create_dataset
    return client


def generate_uuid():
    idx = 0
    while True:
        yield uuid.UUID(int=idx)
        idx += 1



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import traceback

from airflow.contrib.operators.bigquery_operator import BigQueryOperator
from google.cloud import bigquery

from marquez_airflow.extractors import (
    BaseExtractor,
    StepMetadata,
    Source,
    Dataset
)
from marquez_airflow.models import DbTableSchema, DbColumn
from marquez_airflow.extractors.sql.experimental.parser import SqlParser
from marquez_airflow.utils import (
    get_job_name
)

# BIGQUERY DAGs doesn't use this.
# Required to pass marquez server validation.
_BIGQUERY_CONN_URL = \
    'jdbc:bigquery://https://www.googleapis.com/bigquery/v2:443'

log = logging.getLogger(__name__)


class BigQueryExtractor(BaseExtractor):
    operator_class = BigQueryOperator

    def __init__(self, operator):
        super().__init__(operator)

    def _source(self) -> Source:
        conn_id = self.operator.bigquery_conn_id
        return Source(
            type="BIGQUERY",
            name=conn_id,
            connection_url=_BIGQUERY_CONN_URL)

    def extract(self) -> [StepMetadata]:
        return []

    def _bq_table_name(self, bq_table):
        project = bq_table.get('projectId')
        dataset = bq_table.get('datasetId')
        table = bq_table.get('tableId')
        return f"{project}.{dataset}.{table}"

    def extract_on_complete(self, task_instance) -> [StepMetadata]:
        log.debug(f"extract_on_complete({task_instance})")
        context = self.parse_sql_context()
        source = self._source()

        try:
            bigquery_job_id = self._get_xcom_bigquery_job_id(task_instance)
            context['bigquery.job_id'] = bigquery_job_id
            if bigquery_job_id is None:
                raise Exception("Xcom could not resolve BigQuery job id")
        except Exception as e:
            log.error(f"Cannot retrieve job details from BigQuery.Client. {e}",
                      exc_info=True)
            context['bigquery.extractor.client_error'] = \
                f"{e}: {traceback.format_exc()}"
            return [StepMetadata(
                name=get_job_name(task=self.operator),
                context=context,
                inputs=None,
                outputs=None
            )]

        inputs = None
        outputs = None
        try:
            client = bigquery.Client()
            try:
                job = client.get_job(job_id=bigquery_job_id)
                job_properties_str = json.dumps(job._properties)
                context['bigquery.job_properties'] = job_properties_str

                inputs = self._get_input_from_bq(job, context, source, client)
                outputs = self._get_output_from_bq(job, source, client)
            finally:
                # Ensure client has close() defined, otherwise ignore.
                # NOTE: close() was introduced in python-bigquery v1.23.0
                if hasattr(client, "close"):
                    client.close()
        except Exception as e:
            log.error(f"Cannot retrieve job details from BigQuery.Client. {e}",
                      exc_info=True)
            context['bigquery.extractor.error'] = \
                f"{e}: {traceback.format_exc()}"

        return [StepMetadata(
            name=get_job_name(task=self.operator),
            inputs=inputs,
            outputs=outputs,
            context=context
        )]

    def _get_input_from_bq(self, job, context, source, client):
        bq_input_tables = job._properties.get('statistics')\
            .get('query')\
            .get('referencedTables')

        input_table_names = [
            self._bq_table_name(bq_t) for bq_t in bq_input_tables
        ]
        try:
            return [
                Dataset.from_table_schema(
                    source=source,
                    table_schema=table_schema
                )
                for table_schema in self._get_table_schemas(
                    input_table_names, client
                )
            ]
        except Exception as e:
            log.warn(f'Could not extract schema from bigquery. {e}')
            context['bigquery.extractor.bq_schema_error'] = \
                f'{e}: {traceback.format_exc()}'
            return [
                Dataset.from_table(source, table)
                for table in input_table_names
            ]

    def _get_output_from_bq(self, job, source, client):
        bq_output_table = job._properties.get('configuration') \
            .get('query') \
            .get('destinationTable')
        output_table_name = self._bq_table_name(bq_output_table)
        table_schema = self._get_table_safely(output_table_name, client)
        if table_schema:
            return [Dataset.from_table_schema(
                source=source,
                table_schema=table_schema
            )]
        else:
            log.warn("Could not resolve output table from bq")
            return [
                Dataset.from_table(source, output_table_name)
            ]

    def _get_table_safely(self, output_table_name, client):
        try:
            return self._get_table(output_table_name, client)
        except Exception as e:
            log.warn(f'Could not extract output schema from bigquery. {e}')
        return None

    def _get_table_schemas(self, tables: [str], client: bigquery.Client) \
            -> [DbTableSchema]:
        # Avoid querying postgres by returning an empty array
        # if no tables have been provided.
        if not tables:
            return []

        return [self._get_table(table, client) for table in tables]

    def _get_table(self, table: str, client: bigquery.Client) -> DbTableSchema:
        bq_table = client.get_table(table)
        if not bq_table._properties:
            return
        table = bq_table._properties

        if not table.get('schema') or not table.get('schema').get('fields'):
            return

        fields = table.get('schema').get('fields')
        columns = [DbColumn(
            name=fields[i].get('name'),
            type=fields[i].get('type'),
            description=fields[i].get('description'),
            ordinal_position=i
        ) for i in range(len(fields))]

        return DbTableSchema(
            schema_name=table.get('tableReference').get('projectId') + '.' +
            table.get('tableReference').get('datasetId'),
            table_name=table.get('tableReference').get('tableId'),
            columns=columns
        )

    def _get_xcom_bigquery_job_id(self, task_instance):
        bigquery_job_id = task_instance.xcom_pull(
            task_ids=task_instance.task_id, key='job_id')

        log.info(f"bigquery_job_id: {bigquery_job_id}")

        return bigquery_job_id

    def parse_sql_context(self):
        context = {
            'sql': self.operator.sql,
        }
        try:
            sql_meta = SqlParser.parse(self.operator.sql)
            log.debug(f"bigquery sql parsed and obtained meta: {sql_meta}")
            context['bigquery.sql.parsed.inputs'] = json.dumps(
                sql_meta.in_tables)
            context['bigquery.sql.parsed.outputs'] = json.dumps(
                sql_meta.out_tables)
        except Exception as e:
            log.error(f"Cannot parse sql query. {e}",
                      exc_info=True)
            context['bigquery.extractor.sql_parser_error'] = \
                f'{e}: {traceback.format_exc()}'
        return context



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import sqlparse
from sqlparse.sql import T
from sqlparse.sql import TokenList

from marquez_airflow.extractors.sql.experimental import SqlMeta

log = logging.getLogger(__name__)


def _is_in_table(token):
    return _match_on(token, [
        'FROM',
        'INNER JOIN',
        'JOIN',
        'FULL JOIN',
        'FULL OUTER JOIN',
        'LEFT JOIN',
        'LEFT OUTER JOIN',
        'RIGHT JOIN',
        'RIGHT OUTER JOIN'
    ])


def _is_out_table(token):
    return _match_on(token, ['INTO'])


def _match_on(token, keywords):
    return token.match(T.Keyword, values=keywords)


def _get_table(tokens, idx):
    idx, token = tokens.token_next(idx=idx)
    table = next(token.flatten()).value
    return idx, table


class SqlParser:
    """
    This class parses a SQL statement.
    """

    @staticmethod
    def parse(sql):
        if sql is None:
            raise ValueError("A sql statement must be provided")

        # Tokenize the SQL statement
        statements = sqlparse.parse(sql)

        # We assume only one statement in SQL
        tokens = TokenList(statements[0].tokens)
        log.debug(f"Tokenized {sql}: {tokens}")

        in_tables = []
        out_tables = []

        idx, token = tokens.token_next_by(t=T.Keyword)
        while token:
            if _is_in_table(token):
                idx, in_table = _get_table(tokens, idx)
                in_tables.append(in_table)
            elif _is_out_table(token):
                idx, out_table = _get_table(tokens, idx)
                out_tables.append(out_table)

            idx, token = tokens.token_next_by(t=T.Keyword, idx=idx)

        return SqlMeta(in_tables, out_tables)
